if initialized == nil then initialized = false; end

if initialized == false then
    initialized = true;
    -- Force cache refresh to ensure we have the latest files
    __loaded = {};
    -- NOTE: Using GameTextGetXX may be wastey. Consider caching some of the ui elements like [?] and [X] and only update them when the menu is clicked
    print("[spell lab] setting up GUI");
    dofile_once( "data/scripts/gun/gun.lua" );
    dofile_once( "mods/spell_lab/files/helper.lua");
    dofile_once( "mods/spell_lab/files/lib/helper.lua");
    dofile_once( "mods/spell_lab/files/lib/wands.lua");
    dofile_once( "data/scripts/lib/utilities.lua" );
    dofile_once( "mods/spell_lab/files/content/challenges.lua" );

    local gui = gui or GuiCreate();
    local full_screen_width, full_screen_height = GuiGetScreenDimensions( gui );
    GuiStartFrame( gui );
    local screen_width, screen_height = GuiGetScreenDimensions( gui );
    local gui_id = 0;
    local id_offset = 0;
    local mod_settings_id = "spell_lab";
    local version = "c5"
    local refresh_actions = false;

    local function previous_data( gui )
        local left_click,right_click,hover,x,y,width,height,draw_x,draw_y,draw_width,draw_height = GuiGetPreviousWidgetInfo( gui );
        if left_click == 1 then left_click = true; elseif left_click == 0 then left_click = false; end
        if right_click == 1 then right_click = true; elseif right_click == 0 then right_click = false; end
        if hover == 1 then hover = true; elseif hover == 0 then hover = false; end
        return left_click,right_click,hover,x,y,width,height,draw_x,draw_y,draw_width,draw_height;
    end

    local type_order = {
        ACTION_TYPE_MODIFIER,
        ACTION_TYPE_PROJECTILE,
        ACTION_TYPE_STATIC_PROJECTILE,
        ACTION_TYPE_OTHER,
        ACTION_TYPE_MATERIAL,
        ACTION_TYPE_DRAW_MANY,
        ACTION_TYPE_UTILITY,
        ACTION_TYPE_PASSIVE,
    };

    local type_text = {
        [ACTION_TYPE_MODIFIER]="Modifier",
        [ACTION_TYPE_PROJECTILE]="Projectile",
        [ACTION_TYPE_STATIC_PROJECTILE]="Static Projectile",
        [ACTION_TYPE_OTHER]="Other",
        [ACTION_TYPE_MATERIAL]="Material",
        [ACTION_TYPE_DRAW_MANY]="Draw",
        [ACTION_TYPE_UTILITY]="Utility",
        [ACTION_TYPE_PASSIVE]="Passive",
    }

    local font = {
        sprite_filepath = "mods/spell_lab/files/font/small_font.xml",
        character_width = 11,
        character_height = 12,
        character_spacing = -7,
        character_specific_spacing = {
            ["W"] = {1,-6},
            ["M"] = {1,-6},
            ["N"] = {1,-7},
            ["_"] = {1,-6},
            ["#"] = {1,-6},
            ["%"] = {1,-6},
            ["/"] = {1,-6},
            ["\\"] = {1,-6},
            ["~"] = {1,-6},
            [","] = {-1,-8},
        }
    }

    function draw_text( gui, font, text, x, y, scale_x, scale_y, alpha, emissive, z_index, horizontal_alignment, vertical_alignment )
        if scale_x == nil then scale_x = 1; end
        if scale_y == nil then scale_y = 1; end
        if alpha == nil then alpha = 1; end
        if emissive == nil then emissive = true; end
        if horizontal_alignment == nil then horizontal_alignment = 0; end
        if vertical_alignment == nil then vertical_alignment = 0; end
        local cx = x;
        local cy = y;
        local data = {};
        for i = 1, #text do
            local c = text:sub(i,i):upper();
            local spacing_before = 0;
            local spacing_after = font.character_spacing;
            local spacing = font.character_specific_spacing[c];
            if spacing ~= nil then
                spacing_before = spacing[1];
                spacing_after = spacing[2];
            end
            cx = cx + (spacing_before) * scale_x;
            --local s = draw_sprite( font.sprite_filepath, cx, cy, 0, 0, scale_x, scale_y, 0, alpha, emissive, z_index, c );
            --table.insert( sprites, s );
            table.insert( data, { rect_animation = c, x = cx, y = cy } );
            cx = cx + (font.character_width + spacing_after) * scale_x;
        end
        local offset_x = 0;
        local offset_y = 0;
        if horizontal_alignment ~= 0 or vertical_alignment ~= 0 then   
            offset_x = -( cx - x ) * horizontal_alignment;
            offset_y = -( font.character_height * scale_y ) * vertical_alignment;
        end
        for _,character in pairs( data or {} ) do
            GuiImage( gui, next_id(), character.x + offset_x, character.y + offset_y, font.sprite_filepath, alpha, scale_x, scale_y, 0, GUI_RECT_ANIMATION_PLAYBACK.Loop, character.rect_animation );
        end
    end

    local action_count = action_count;
    local sorted_actions = sorted_actions or {};
    local action_data = action_data or {};
    local action_metadata = action_metadata or {};
    local filter_type = filter_type or ACTION_TYPE_PROJECTILE;
    local selected_acton_index = 0;
    local current_challenge = nil;
    local action_spell_filter = "";
    local is_panel_open = true;

    function get_action_metadata( action_id )
        if action_metadata[action_id] == nil and action_data[action_id] then
            local metadata = {
                c = {},
                projectiles = nil,
                shot_effects = {},
            };
            local skip_values = {
                damage_by_type = true,
                mTriggers = true,
                config = true,
                config_explosion = true,
                damage_critical = true,
            };
            reflecting = true;
            Reflection_RegisterProjectile = function( filepath )
                metadata.projectiles = metadata.projectiles or {};
                if metadata.projectiles[filepath] == nil then
                    local projectile_entity = EntityCreateNew();
                    EntityApplyTransform( projectile_entity, 14600, -60000 );
                    EntityLoadToEntity( filepath, projectile_entity );
                    local projectile = EntityGetFirstComponent( projectile_entity, "ProjectileComponent" );
                    if projectile then
                        metadata.projectiles[filepath] = {};
                        for k,v in pairs( ComponentGetMembers( projectile ) ) do
                            if skip_values[k] ~= true then
                                metadata.projectiles[filepath][k] = ComponentGetValue2( projectile, k );
                            end
                        end
                        metadata.projectiles[filepath].projectiles = 1;
                        metadata.projectiles[filepath].lifetime = ComponentGetValue2( projectile, "mStartingLifetime" );
                        ComponentSetValue2( projectile, "on_death_explode", false );
                        ComponentSetValue2( projectile, "on_lifetime_out_explode", false );
                        ComponentSetValue2( projectile, "collide_with_entities", false );
                        ComponentSetValue2( projectile, "collide_with_world", false );
                        ComponentSetValue2( projectile, "lifetime", 999 );

                        local damage_by_type = ComponentObjectGetMembers( projectile, "damage_by_type" );
                        if damage_by_type then
                            for type,value in pairs( damage_by_type or {} ) do
                                metadata.projectiles[filepath]["damage_"..type] = ComponentObjectGetValue2( projectile, "damage_by_type", type ) or 0;
                            end
                        end

                        local config_explosion = ComponentObjectGetMembers( projectile, "config_explosion" );
                        if config_explosion then
                            metadata.projectiles[filepath].damage_explosion = ComponentObjectGetValue2( projectile, "config_explosion", "damage" );
                        end
                    end
                    EntityKill( projectile_entity );
                else
                    metadata.projectiles[filepath].projectiles = metadata.projectiles[filepath].projectiles + 1;
                end
            end
            local _shot_effects = shot_effects;
            local _c = c;
                c = {};
                shot_effects = {};
                current_reload_time = 0;
                local draws = 0;
                local _draw_actions = draw_actions;
                draw_actions = function( how_many ) draws = draws + how_many; end
                reset_modifiers( c );
                ConfigGunShotEffects_Init( shot_effects );
                action_data[action_id].action();
                draw_actions = _draw_actions;
                c.draw_actions = draws;
                c.reload_time = current_reload_time;
                c.mana = action_data[action_id].mana;
                metadata.c = c;
            c = _c;
            reflecting = false;
            action_metadata[action_id] = metadata;
        end
        return action_metadata[action_id];
    end

    local FORMAT = {
        Floor = 0,
        Round = 1,
        Ceiling = 2
    }

    local function format_value( value, decimals, show_sign, format )
        local text = "";
        if value ~= nil then
            if show_sign == true and value > 0 then
                text = "+";
            end
            local rounder = math.floor;
            local value_offset = 0;
            if format == FORMAT.Ceiling then
                rounder = math.ceil;
            elseif format == FORMAT.Round then
                value_offset = 0.5;
            end
            local power = math.pow( 10, decimals );
            text = text .. tostring( rounder( value * power + value_offset ) / power );
        else
            return "missing"
        end
        return text;
    end

    local metadata_to_show = {
        c = {
            draw_actions = { "Draw", 0, function(value) return format_value( value, 0 ); end },
            max_uses = { "Max Uses", nil, function(value) return format_value( value, 0 ); end },
            mana = { "Mana Drain", 0, function(value) return format_value( value, 0 ); end },
            fire_rate_wait = { "Cast Delay", 0, function(value) return format_value( value / 60, 3, true ) .. " s"; end },
            speed_multiplier = { "Proj. Speed", 1, function(value) return "x " .. format_value( value, 2 ); end },
            reload_time = { "Recharge Time", 0, function(value) return format_value( value / 60, 3, true ) .. " s"; end },
            spread_degrees = { "Spread", 0, function(value) return format_value( value, 2, true, FORMAT.Round ) .. " DEG"; end },
            damage_critical_chance = { "Crit. Chance", 0, function(value) return format_value( value, 0 ) .. "%"; end },
            explosion_radius = { "Expl. Radius", 0, function(value) return format_value( value, 0, true, FORMAT.Ceiling ); end },
            damage_projectile_add = { "Dmg. (Projectile)", 0, function(value) return format_value( value * 25, 2, true, FORMAT.Ceiling ); end },
            damage_explosion_add = { "Dmg. (Explosion)", 0, function(value) return format_value( value * 25, 2, true, FORMAT.Ceiling ); end },
            damage_melee_add = { "Dmg. (Melee)", 0, function(value) return format_value( value * 25, 2, true, FORMAT.Ceiling ); end },
            damage_electricity_add = { "Dmg. (Electric)", 0, function(value) return format_value( value * 25, 2, true, FORMAT.Ceiling ); end },
            damage_fire_add = { "Dmg. (Fire)", 0, function(value) return format_value( value * 25, 2, true, FORMAT.Ceiling ); end },
        },
        projectiles = {
            { "Projectiles", function( data ) if data.projectiles ~= nil and data.projectiles ~= 0 then return format_value( data.projectiles, 0 ); end end },
            { "Dmg. (Projectile)", function( data ) if data.damage ~= nil and data.damage ~= 0 then return format_value( data.damage * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Explosion)", function( data ) if data.damage_explosion ~= nil and data.damage_explosion ~= 0 then return format_value( data.damage_explosion * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Slice)", function( data ) if data.damage_slice ~= nil and data.damage_slice ~= 0 then return format_value( data.damage_slice * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Melee)", function( data ) if data.damage_melee ~= nil and data.damage_melee ~= 0 then return format_value( data.damage_melee * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Electric)", function( data ) if data.damage_electricity ~= nil and data.damage_electricity ~= 0 then return format_value( data.damage_electricity * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Fire)", function( data ) if data.damage_fire ~= nil and data.damage_fire ~= 0 then return format_value( data.damage_fire * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Ice)", function( data ) if data.damage_ice ~= nil and data.damage_ice ~= 0 then return format_value( data.damage_ice * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Drill)", function( data ) if data.damage_drill ~= nil and data.damage_drill ~= 0 then return format_value( data.damage_drill * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Poison)", function( data ) if data.damage_poison ~= nil and data.damage_poison ~= 0 then return format_value( data.damage_poison * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Healing)", function( data ) if data.damage_healing ~= nil and data.damage_healing ~= 0 then return format_value( data.damage_healing * 25, 2, false, FORMAT.Round ); end end },
            { "Dmg. (Radioactive)", function( data ) if data.damage_radioactive ~= nil and data.damage_radioactive ~= 0 then return format_value( data.damage_radioactive * 25, 2, false, FORMAT.Round ); end end },
            { "Speed", function( data )
                if data.speed_min ~= nil and data.speed_max ~= nil then
                    if data.speed_min ~= data.speed_max then
                        return tostring( data.speed_min.."-"..data.speed_max );
                    elseif data.speed_min ~= 0 then
                        return tostring( data.speed_min );
                    end
                end
            end },
            { "Spread (Innate)", function( data ) if data.direction_random_rad ~= 0 then return format_value( data.direction_random_rad * 180 / math.pi, 1, false, FORMAT.Round ) .. " DEG"; end end },
            -- NOTE Lifetime is randomized so this won't be accurate
            --{ "Lifetime", function( data ) return tostring( (data.lifetime_randomness == 0 and data.lifetime) or (data.lifetime - data.lifetime_randomness) .. "-" .. (data.lifetime + data.lifetime_randomness ) ); end },
        }
    };

    function get_player_or_camera_position()
        local x, y = GameGetCameraPos();
        local player = EntityGetWithTag( "player_unit" )[1];
        if player then
            x, y = EntityGetTransform( player );
        end
        return x, y;
    end

    function get_held_wand( do_random )
        if do_random == nil then do_random = false; end
        local player = EntityGetWithTag( "player_unit" )[1];
        if player then
            local held_wand = get_entity_held_or_random_wand( player, do_random );
            if held_wand then
                return held_wand;
            end
        end
    end

    function get_screen_position( x, y )
        local screen_width, screen_height = GuiGetScreenDimensions( gui );
        local camera_x, camera_y = GameGetCameraPos();
        local res_width = MagicNumbersGetValue( "VIRTUAL_RESOLUTION_X" );
        local res_height = MagicNumbersGetValue( "VIRTUAL_RESOLUTION_Y" );
        local ax = (x - camera_x) / res_width * screen_width;
        local ay = (y - camera_y) / res_height * screen_height;
        return ax + screen_width * 0.5, ay + screen_height * 0.5;
    end

    function force_refresh_held_wands()
        local player = EntityGetWithTag( "player_unit" )[1];
        if player then
            local inventory2 = EntityGetFirstComponent( player, "Inventory2Component" );
            if inventory2 ~= nil then
                ComponentSetValue2( inventory2, "mForceRefresh", true );
                ComponentSetValue2( inventory2, "mActualActiveItem", 0 );
            end
        end
    end

    function remove_action( action_entity, wand )
        EntityRemoveFromParent( action_entity );
        EntityKill( action_entity );
        if wand then
            local item = EntityGetFirstComponentIncludingDisabled( action_entity, "ItemComponent" );
            if item then
                if ComponentGetValue2( item, "permanently_attached" ) then
                    wand_adjust_stat( wand, "deck_capacity", function( value ) return value - 1; end );
                end
            end
            force_refresh_held_wands();
        end
    end

    function toggle_always_cast( action_entity, wand )
        local item = EntityGetFirstComponentIncludingDisabled( action_entity, "ItemComponent" );
        if item then
            if ComponentGetValue2( item, "is_frozen" ) == false then
                if ComponentGetValue2( item, "permanently_attached" ) then
                    local x,y = ComponentGetValue2( item, "inventory_slot" );
                    wand_explode_action( wand, x, false ); 
                    ComponentSetValue2( item, "permanently_attached", false );
                    if wand then wand_adjust_stat( wand, "deck_capacity", function( value ) return value - 1; end ) end
                else
                    ComponentSetValue2( item, "permanently_attached", true );
                    if wand then wand_adjust_stat( wand, "deck_capacity", function( value ) return value + 1; end ) end
                end
            else
                GamePrint( "Can't modify a locked spell" );
            end
        end
    end

    function is_action_unlocked( action )
        local valid = false;
        if action then
            if action.spawn_requires_flag ~= nil then
                local flag_name = action.spawn_requires_flag;
                local flag_status = HasFlagPersistent( flag_name );
                if flag_status then valid = true; end
                -- Music Notes
                --if action.spawn_probability == "0" then  valid = false; end
            else
                valid = true;
            end
        end
        return valid;
    end

    function toggle_all_action_unlocks( which )
        for _,action in pairs(actions) do
            if action.spawn_requires_flag then
                if which == true then
                    if not HasFlagPersistent( action.spawn_requires_flag ) then
                        AddFlagPersistent( action.spawn_requires_flag );
                    end
                else
                    if HasFlagPersistent( action.spawn_requires_flag ) then
                        RemoveFlagPersistent( action.spawn_requires_flag );
                    end
                end
            end
        end
    end

    function acquire_action( action_id, target_x, do_insert, force )
        if do_insert == nil then do_insert = false; end
        local player = EntityGetWithTag( "player_unit" )[1];
        if player then
            local held_wand = get_held_wand();
            if held_wand then
                local action_entity = CreateItemActionEntity( action_id, x, y );
                local added_to_wand = false;
                if ModSettingGet( "spell_lab.quick_spell_picker" ) == true and held_wand ~= 0 then
                    local held_wand_actions = wand_get_actions_absolute( held_wand );
                    local ability = EntityGetFirstComponentIncludingDisabled( held_wand, "AbilityComponent" );
                    if ability then
                        local max_actions = ability_component_get_stat( ability, "deck_capacity" );
                        local used_slots = {};
                        for k,v in pairs(held_wand_actions) do
                            used_slots[k] = true;
                        end
                        local next_free_slot = nil;
                        for i=0,max_actions do
                            if used_slots[i] ~= true then
                                next_free_slot = i;
                                break;
                            end
                        end
                        if target_x == nil then
                            target_x = next_free_slot;
                        end
                        local abandon_adding = false;
                        if not do_insert then
                            if held_wand_actions[target_x] then
                                if held_wand_actions[target_x].locked ~= true then
                                    EntityRemoveFromParent( held_wand_actions[target_x].entity );
                                    EntityKill( held_wand_actions[target_x].entity );
                                else
                                    GamePrint("Can't remove a locked spell");
                                    abandon_adding = true;
                                end
                            end
                        elseif next_free_slot < max_actions then
                            for i=target_x,next_free_slot-1 do
                                ComponentSetValue2( held_wand_actions[i].item, "inventory_slot", held_wand_actions[i].x + 1, held_wand_actions[i].y );
                            end
                        else
                            abandon_adding = true;
                        end
                        if not abandon_adding and target_x < max_actions then
                            EntityAddChild( held_wand, action_entity );
                            local item = EntityGetFirstComponentIncludingDisabled( action_entity, "ItemComponent" );
                            local item_x, item_y = ComponentGetValue2( item, "inventory_slot" );
                            ComponentSetValue2( item, "inventory_slot", target_x, item_y );
                            EntitySetComponentsWithTagEnabled( action_entity, "enabled_in_world", false );
                            force_refresh_held_wands();
                            added_to_wand = true;
                        end
                        selected_acton_index = ( selected_acton_index + 1 ) % max_actions;
                    end
                end
                local added_to_inventory = false;
                if not added_to_wand then
                    local full_inventory = nil;
                    local player_child_entities = EntityGetAllChildren( player );
                    if player_child_entities ~= nil then
                        for i,child_entity in ipairs( player_child_entities ) do
                            if EntityGetName( child_entity ) == "inventory_full" then
                                full_inventory = child_entity;
                                break;
                            end
                        end
                    end

                    -- set inventory contents
                    if full_inventory ~= nil then
                        EntitySetComponentsWithTagEnabled( action_entity, "enabled_in_world", false );
                        EntityAddChild( full_inventory, action_entity );
                        GamePrint( GameTextGetTranslatedOrNot(action_data[action_id].name) .. " added to your inventory" );
                        added_to_inventory = true;
                    end
                end
            end
        end
    end

    local PICKERS = {
        None = nil,
        SpellPicker = "spell_lab_spell_picker",
        ChallengePicker = "spell_lab_challenge_picker",
        WandPicker = "spell_lab_wand_picker",
        WandBox = "spell_lab_wand_box",
    };

    local active_picker = nil;

    function is_picker_active( picker )
        return active_picker == picker;
    end

    function change_picker( picker )
        if active_picker == picker then
            active_picker = nil;
        else
            active_picker = picker;
            if active_picker == PICKERS.SpellPicker then
                refresh_actions = true;
            end
        end
    end

    function collect_action_data()
        sorted_actions = {};
        local player = EntityGetWithTag( "player_unit" )[1];
        if player then EntityRemoveTag( player, "player_unit" ); end
        action_count = 0;
        for k,action in pairs( actions ) do
            action_count = (action_count or 0) + 1;
            if is_action_unlocked( action ) then
                sorted_actions[action.type] = (sorted_actions[action.type] or {});
                table.insert( sorted_actions[action.type], action );
                action_data[action.id] = action;
                get_action_metadata( action.id );
            end
        end
        for k,v in pairs( sorted_actions ) do
            table.sort( v, function( a, b )
                if a.type == b.type then
                    return GameTextGetTranslatedOrNot( a.name ):lower() < GameTextGetTranslatedOrNot( b.name ):lower();
                else
                    return a.type < b.type;
                end
            end);
        end
        if player then EntityAddTag( player, "player_unit" ); end
        refresh_actions = false;
    end

    collect_action_data();

    if preset_wands == nil then preset_wands = dofile( "mods/spell_lab/files/wand_presets.lua"); end

    if saved_wands == nil then
        saved_wands = {};
        local smallfolk = dofile( "mods/spell_lab/files/lib/smallfolk.lua" );
        local saved_wand_data = ModSettingGet( "spell_lab_saved_wands" );
        if saved_wand_data ~= nil then
            local loaded_wands = smallfolk.loads( saved_wand_data );
            for k,v in pairs(loaded_wands) do
                table.insert( saved_wands, v );
            end
        end
    end

    function serialize_saved_wands()
        local smallfolk = dofile( "mods/spell_lab/files/lib/smallfolk.lua" );
        if saved_wands ~= nil then
            ModSettingSet( "spell_lab_saved_wands", smallfolk.dumps( saved_wands ) );
        end
    end

    local word_wrap = function( str, wrap_size )
        if wrap_size == nil then wrap_size = 60; end
        local last_space_index = 1;
        local last_wrap_index = 0;
        for i=1,#str do
            if str:sub(i,i) == " " then
                last_space_index = i;
            end
            if str:sub(i,i) == "\n" then
                last_space_index = i;
                last_wrap_index = i;
            end
            if i - last_wrap_index > wrap_size then
                str = str:sub(1,last_space_index-1) .. "\n" .. str:sub(last_space_index + 1);
                last_wrap_index = i;
            end
        end
        return str;
    end

    local truncate_long_string = function( str )
        if #str > 31 then
            str = str:sub( 0, 14 ).."..."..str:sub( -14 );
        end
        return str;
    end

    function next_id()
        id_offset = id_offset + 1;
        return gui_id + id_offset;
    end

    local wand_stats = {
        { 
            stat = "shuffle_deck_when_empty",
            label = "Shuffle",
            default = 0, current = 0,
            min = 0, max = 1,
            formatter = " ",
            value_callback = function( value )
                if type( value ) == "boolean" then
                    if value == true then
                        value = 1;
                    else
                        value = 0;
                    end
                end
                return math.floor( value + 0.5 );
            end,
            text_callback = function( value ) return value == 0 and "No" or "Yes"; end,
        },
        {  
            stat = "actions_per_round",
            label = "Spells/Cast",
            default = 1, current = 1,
            min = 1, max = 26,
            formatter = " ",
            value_callback = function( value ) return math.floor( value + 0.5 ); end,
            text_callback = function( value ) return tostring( value ); end,
        },   
        {
            stat = "fire_rate_wait",
            label = "Cast delay",
            default = 10, current = 10,
            min = -21, max = 240,
            formatter = " ",
            value_callback = function( value ) return math.floor( value + 0.5 ); end,
            text_callback = function( value ) return ( math.floor( value / 60 * 100 + 0.5 ) / 100 ) .. " s"; end,
        },
        {
            stat = "reload_time",
            label = "Rechrg. Time",
            default = 20, current = 20,
            min = -21, max = 240,
            formatter = " ",
            value_callback = function( value ) return math.floor( value + 0.5 ); end,
            text_callback = function( value ) return ( math.floor( value / 60 * 100 + 0.5 ) / 100 ) .. " s"; end,
        },
        {
            stat = "mana_max",
            label = "Mana max",
            default = 100, current = 100,
            min = 0, max = 5000,
            formatter = " ",
            value_callback = function( value ) return math.floor( value + 0.5 ); end,
            text_callback = function( value ) return tostring( value ); end,
        },
        {
            stat = "mana_charge_speed",
            label = "Mana chg. spd",
            default = 30, current = 30,
            min = 0, max = 5000,
            formatter = " ",
            value_callback = function( value ) return math.floor( value + 0.5 ); end,
            text_callback = function( value ) return tostring( value ); end,
        },
        {  
            stat = "deck_capacity",
            label = "Capacity",
            default = 3, current = 3,
            min = 1, max = 26,
            formatter = " ",
            value_callback = function( value ) return math.floor( value + 0.5 ); end,
            text_callback = function( value ) return tostring( value ); end,
        },
        {
            stat = "spread_degrees",
            label = "Spread",
            default = 0, current = 0,
            min = -30, max = 30,
            formatter = " ",
            value_callback = function( value ) return math.floor( value + 0.5 ); end,
            text_callback = function( value ) return ( math.floor( value * 10 ) / 10 ) .. " DEG" end,
        },
        {
            stat = "speed_multiplier",
            label = "Speed multiplier",
            default = 1.0, current = 1.0,
            min = 0.0, max = 2.0,
            formatter = " ",
            value_callback = function( value ) return math.floor( value * 1000 ) / 1000; end,
            text_callback = function( value ) return ( math.floor( value * 1000 ) / 1000 .. " x" ) end,
        }
    }

    function do_picker_button( filepath, picker, option_text )
        if not is_picker_active( picker ) then
            GuiOptionsAddForNextWidget( gui, GUI_OPTION.DrawSemiTransparent );
        end
        if GuiImageButton( gui, next_id(), 0, 0, "", filepath ) then
            change_picker( picker );
        end
        if is_picker_active( picker ) then
            GuiTooltip( gui, "Disable "..option_text, "" );
        else
            GuiTooltip( gui, "Enable "..option_text, "" );
        end
    end

    function do_flag_toggle_image_button( filepath, flag, option_text )
        if ModSettingGet( mod_settings_id.."."..flag ) ~= true then
            GuiOptionsAddForNextWidget( gui, GUI_OPTION.DrawSemiTransparent );
        end
        if GuiImageButton( gui, next_id(), 0, 0, "", filepath ) then
            if ModSettingGet( mod_settings_id.."."..flag ) == true then
                ModSettingSet( mod_settings_id.."."..flag, false );
            else
                ModSettingSet( mod_settings_id.."."..flag, true );
            end
        end
        if ModSettingGet( mod_settings_id.."."..flag ) == true then
            GuiTooltip( gui, "Disable "..option_text, "" );
        else
            GuiTooltip( gui, "Enable "..option_text, "" );
        end
    end

    -- TODO fix this when you learn how to
    function do_action_button( action_id, x, y, selected, click_callback, verbose_tooltip, permanent, note )
        if x == nil then x = 0; end
        if y == nil then y = 0; end
        local image_sprite = "mods/spell_lab/files/gui/buttons/empty_spell.png";
        local this_action_data =  action_data[action_id];
        local left_click,right_click,hover;
        local spell_box_suffix = "";
        
        if this_action_data then spell_box_suffix = spell_box_suffix .. "_" .. this_action_data.type; end
        if this_action_data then
            image_sprite = this_action_data.sprite;
        end
        local this_action_metadata = get_action_metadata( action_id );

        GuiOptionsAddForNextWidget( gui, GUI_OPTION.HandleDoubleClickAsClick );
        GuiOptionsAddForNextWidget( gui, GUI_OPTION.ClickCancelsDoubleClick );
        GuiImageButton( gui, next_id(), 2, 2, "", image_sprite );
        left_click,right_click,hover = previous_data( gui );

        if selected == true then
                spell_box_suffix = spell_box_suffix .. "_active";
        elseif hover then
            spell_box_suffix = spell_box_suffix .. "_hover";
        end

        --if tooltip_string ~= nil then GuiTooltip( gui, tooltip_string, "" ); end
        --if this_action_data then GuiTooltip( gui, GameTextGetTranslatedOrNot( this_action_data.name ) .. " (" .. this_action_data.id .. ")", "" ); end
        if this_action_data then
            do_custom_tooltip( function()
                GuiLayoutBeginVertical( gui, 0, 0 );
                    GuiText( gui, 0, 0, GameTextGetTranslatedOrNot( this_action_data.name ) );
                    if verbose_tooltip == true then
                        GuiColorSetForNextWidget( gui, 0.5, 0.5, 0.5, 1.0 );
                        GuiText( gui, 0, 0, this_action_data.id );
                        GuiColorSetForNextWidget( gui, 0.5, 0.5, 1.0, 1.0 );
                        GuiText( gui, 0, 0, GameTextGetTranslatedOrNot( type_text[ this_action_data.type ] ) );
                        GuiColorSetForNextWidget( gui, 0.811, 0.811, 0.811, 1.0 );
                        GuiText( gui, 0, 0, word_wrap( GameTextGetTranslatedOrNot( this_action_data.description ) ) );
                    end
                    if note then
                        GuiColorSetForNextWidget( gui, 0.5, 0.5, 0.5, 1.0 );
                        GuiText( gui, 0, 0, note );
                    end
                    if this_action_metadata and verbose_tooltip then
                        GuiLayoutAddVerticalSpacing( gui, 5 );
                        for p_key,d in pairs( metadata_to_show.projectiles ) do
                            if this_action_metadata.projectiles then
                                for filepath,data in pairs( this_action_metadata.projectiles ) do
                                    local value = d[2]( data );
                                    if value and value ~= d[3] then
                                        local text = tostring( value );
                                        GuiLayoutBeginHorizontal( gui, 0, 0 );
                                            GuiColorSetForNextWidget( gui, 0.811, 0.811, 0.811, 1.0 );
                                            GuiText( gui, 0, 0, d[1] );
                                            GuiColorSetForNextWidget( gui, 1.0, 0.75, 0.5, 1.0 );
                                            local left_click,right_click,hover,x,y,width,height,draw_x,draw_y,draw_width,draw_height = previous_data( gui );
                                            --GuiOptionsAddForNextWidget( gui, GUI_OPTION.Layout_NoLayouting );
                                            GuiText( gui, 72 - width, 0, tostring( value ) );
                                        GuiLayoutEnd( gui );
                                        GuiLayoutAddVerticalSpacing( gui, -2 );
                                        break;
                                    end
                                end
                            end
                        end
                        for c_key,d in pairs( metadata_to_show.c ) do
                            if this_action_metadata.c[c_key] ~= d[2] then
                                if this_action_metadata.c[c_key] ~= nil then
                                    local text = d[1]..tostring( this_action_metadata.c[c_key] );
                                    if d[3] then
                                        text = tostring( d[3]( this_action_metadata.c[c_key] ) );
                                    end
                                    GuiLayoutBeginHorizontal( gui, 0, 0 );
                                        GuiColorSetForNextWidget( gui, 0.811, 0.811, 0.811, 1.0 );
                                        GuiText( gui, 0, 0, d[1] );
                                        GuiColorSetForNextWidget( gui, 1.0, 0.75, 0.5, 1.0 );
                                        local left_click,right_click,hover,x,y,width,height,draw_x,draw_y,draw_width,draw_height = previous_data( gui );
                                        --GuiOptionsAddForNextWidget( gui, GUI_OPTION.Layout_NoLayouting );
                                        GuiText( gui, 72 - width, 0, text );
                                    GuiLayoutEnd( gui );
                                    GuiLayoutAddVerticalSpacing( gui, -2 );
                                end
                            end
                        end
                    end
                GuiLayoutEnd( gui );
            end, nil, 10 );
        end
        GuiZSet( gui, 0 );

        GuiZSetForNextWidget( gui, 1 );
        GuiImage( gui, next_id(), -20, 0, "mods/spell_lab/files/gui/buttons/spell_box"..spell_box_suffix..".png", 1.0, 1.0, 0 );
        if permanent then
            local left_click,right_click,hover,x,y,width,height,draw_x,draw_y,draw_width,draw_height = previous_data( gui );
            GuiOptionsAddForNextWidget( gui, GUI_OPTION.Layout_NoLayouting );
            GuiImage( gui, next_id(), x-2, y-2, "data/ui_gfx/inventory/icon_gun_permanent_actions.png", 1.0, 1.0, 0 );
        end

        if click_callback and ( left_click or right_click ) then
            click_callback( left_click, right_click );
        end
    end

    function do_custom_tooltip( callback, z, x_offset, y_offset )
        if z == nil then z = -12; end
        local left_click,right_click,hover,x,y,width,height,draw_x,draw_y,draw_width,draw_height = previous_data( gui );
        local screen_width,screen_height = GuiGetScreenDimensions( gui );
        if x_offset == nil then x_offset = 0; end
        if y_offset == nil then y_offset = 0; end
        if draw_y > screen_height * 0.5 then
            y_offset = y_offset - height;
        end
        if hover then
            local screen_width, screen_height = GuiGetScreenDimensions( gui );
            GuiZSet( gui, z );
            GuiLayoutBeginLayer( gui );
                GuiLayoutBeginVertical( gui, ( x + x_offset + width ), ( y + y_offset ), true );
                    GuiBeginAutoBox( gui );
                        if callback ~= nil then callback(); end
                        GuiZSetForNextWidget( gui, z + 1 );
                    GuiEndAutoBoxNinePiece( gui );
                GuiLayoutEnd( gui );
            GuiLayoutEndLayer( gui );
        end
    end

    function do_simple_action_list( absolute_actions, permanent_actions, capacity )   
        for _,permanent_action in pairs( permanent_actions or {} ) do
            local this_action_data = action_data[permanent_action.action_id];
            if this_action_data then GuiImage( gui, next_id(), 0, 0, this_action_data.sprite, 1.0, 0.5, 0 ); end
        end

        if capacity > 0 then
            for i=1,capacity do
                local action = absolute_actions[i-1] or {};
                local this_action_data = action_data[action.action_id];
                if this_action_data then GuiImage( gui, next_id(), 0, 0, this_action_data.sprite, 1.0, 0.5, 0 ); end
            end
        end
    end

    function do_flat_action_list( actions )   
        for _,action_id in ipairs(actions) do
            local this_action_data = action_data[action_id];
            if this_action_data then GuiImage( gui, next_id(), 0, 0, this_action_data.sprite, 1.0, 0.5, 0 ); end
        end
    end

    function do_action_list( absolute_actions, permanent_actions, capacity, wand )
        for _,permanent_action in pairs( permanent_actions or {} ) do
            do_action_button( permanent_action.action_id, 0, 0, false, function( left_click, right_click )
                if permanent_action and right_click then
                    if permanent_action.locked == false then
                        remove_action( permanent_action.entity, wand );
                    elseif permanent_action.locked == true then
                        GamePrint("Can't remove a locked spell");
                    end
                elseif permanent_action and left_click then
                    toggle_always_cast( permanent_action.entity, wand );
                    force_refresh_held_wands();
                end
            end, false, true, "Left Click to demote spell" );
        end

        if capacity > 0 then
            for i=1,capacity do
                local action = absolute_actions[i-1] or {};
                local note = "Left Click to select this spell slot";
                if (selected_acton_index == i - 1) then
                    note = permanent == true and "Left Click to demote spell" or "Left Click to promote spell to Always Cast";
                end

                do_action_button( action.action_id, 0, 0, selected_acton_index == i - 1, function( left_click, right_click )
                    if action and right_click == true then
                        if action.locked == false then
                            remove_action( action.entity, wand );
                        elseif action.locked == true then
                            GamePrint( "Can't remove a locked spell" );
                        end
                    elseif action and left_click and selected_acton_index == i - 1 then
                        toggle_always_cast( action.entity, wand );
                        force_refresh_held_wands();
                    end
                    selected_acton_index = i - 1;
                end, false, false, note );
            end
        end
    end

    function do_wand_stats( gui, stats )
        for k,v in pairs( wand_stats ) do
            if stats[ v.stat ] then
                GuiLayoutBeginHorizontal( gui, 0, 0 );
                    GuiColorSetForNextWidget( gui, 0.811, 0.811, 0.811, 1.0 );
                    GuiText( gui, 0, 0, v.label );
                    GuiColorSetForNextWidget( gui, 1.0, 0.75, 0.5, 1.0 );
                    local left_click,right_click,hover,x,y,width,height,draw_x,draw_y,draw_width,draw_height = previous_data( gui );
                    GuiText( gui, 72 - width, 0, v.text_callback( stats[ v.stat ] ) );
                GuiLayoutEnd( gui );
                GuiLayoutAddVerticalSpacing( gui, -4 );
            end
        end
    end

    function do_gui()
        if refresh_actions == true then
            collect_action_data();
        end
        local now = GameGetFrameNum();
        id_offset = 0;
        GuiStartFrame( gui );
        GuiIdPushString( gui, "spell_lab" );
        GuiOptionsAdd( gui, GUI_OPTION.NoPositionTween );
        if is_panel_open and GameIsInventoryOpen() == false and GlobalsGetValue( "spell_lab_dead", "0" ) == "0" and GameHasFlagRun( "gkbrkn_config_menu_open" ) == false then
            local player = EntityGetWithTag( "player_unit" )[1];
            GuiLayoutBeginVertical( gui, 0, 2 );
                GuiLayoutBeginHorizontal( gui, 42, 2 );
                    if GlobalsGetValue( "spell_lab_checkpoint", "0" ) == "0" then
                        GuiOptionsAddForNextWidget( gui, GUI_OPTION.DrawSemiTransparent );
                    end
                    if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/spell_lab.png" ) then
                        if player then
                            if current_challenge ~= nil then
                                --GamePrintImportant( "Challenge Abandoned", "" );
                                current_challenge = nil;
                            end
                            if GlobalsGetValue( "spell_lab_checkpoint", "0" ) == "0" then
                                local px, py = EntityGetTransform( player );
                                GlobalsSetValue( "spell_lab_checkpoint", math.floor( px )..","..math.floor( py ) );
                                EntityApplyTransform( player, 14600, -6050 );
                                GameSetCameraPos( 14600, -6050 );
                            else
                                local s = string_split( GlobalsGetValue( "spell_lab_checkpoint", "0" ), "," );
                                local cx = tonumber( s[1] );
                                local cy = tonumber( s[2] ) - 10;
                                EntityApplyTransform( player, cx, cy );
                                GameSetCameraPos( cx, cy );
                                GlobalsSetValue( "spell_lab_checkpoint", "0" );
                            end
                        end
                    end
                    if GlobalsGetValue( "spell_lab_checkpoint", "0" ) == "0" then
                        GuiTooltip( gui, "Enter the Spell Lab", "" );
                    else
                        GuiTooltip( gui, "Leave the Spell Lab", "" );
                    end

                    do_picker_button( "mods/spell_lab/files/gui/buttons/edit_wands.png", PICKERS.SpellPicker, "Spell Picker ");
                    do_picker_button( "mods/spell_lab/files/gui/buttons/challenge_picker.png", PICKERS.ChallengePicker, "Challenge Picker" );
                    do_picker_button( "mods/spell_lab/files/gui/buttons/wand_spawner.png", PICKERS.WandPicker, "Wand Spawner" );
                    do_picker_button( "mods/spell_lab/files/gui/buttons/wand_list.png", PICKERS.WandBox, "Wand Box" );
                    do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/show_toggles.png", "show_toggle_options", "Toggle Options" )
                GuiLayoutEnd( gui );

                GuiLayoutBeginHorizontal( gui, 44, 2 );
                    if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/clear_projectiles.png" ) then
                        for k,v in pairs( EntityGetWithTag( "projectile_player" ) or {} ) do
                            local projectile = EntityGetFirstComponent( v, "ProjectileComponent" );
                            if projectile ~= nil then
                                ComponentSetValue2( projectile, "on_death_explode", false );
                                ComponentSetValue2( projectile, "on_lifetime_out_explode", false );
                            end
                            EntityKill(v);
                        end
                    end
                    GuiTooltip( gui, "Clear Projectiles", "" );
                    
                    if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/clear_wait.png" ) then
                        if player then
                            for k,v in pairs( EntityGetWithTag("wand") ) do
                                if EntityGetRootEntity( v ) == player then
                                    local ability = EntityGetFirstComponentIncludingDisabled( v, "AbilityComponent" );
                                    if ability then
                                        ComponentSetValue2( ability, "mReloadFramesLeft", 0 );
                                        ComponentSetValue2( ability, "mNextFrameUsable", now );
                                        ComponentSetValue2( ability, "mReloadNextFrameUsable", now );
                                    end
                                end
                            end
                        end
                    end
                    GuiTooltip( gui, "Clear Cast Delay / Recharge Time", "" );

                    if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/clear_wand.png" ) then
                        if player then
                            local held_wand = get_held_wand();
                            if held_wand ~= 0 then
                                wand_clear_actions( held_wand );
                                force_refresh_held_wands();
                            end
                        end
                    end
                    GuiTooltip( gui, "Clear Held Wand", "" );

                    if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/spawn_target_dummy.png" ) then
                        local x,y = GameGetCameraPos();
                        if player then
                            x,y = EntityGetTransform( player );
                        end
                        EntityLoad( "mods/spell_lab/files/entities/dummy_target/dummy_target.xml", x, y );
                    end
                    GuiTooltip( gui, "Spawn Target Dummy", "" );

                    if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/spawn_boss_target_dummy.png" ) then
                        local x,y = GameGetCameraPos();
                        if player then
                            x,y = EntityGetTransform( player );
                        end
                        EntityLoad( "mods/spell_lab/files/entities/dummy_target/dummy_target_final.xml", x, y );
                    end
                    GuiTooltip( gui, "Spawn Boss Target Dummy", "" );
                GuiLayoutEnd( gui );

                if ModSettingGet("spell_lab.show_toggle_options") == true then
                    GuiLayoutBeginHorizontal( gui, 35, 2 );
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/disable_projectiles.png", "disable_projectiles", "No Projectiles" )
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/disable_particles.png", "no_particles", "No Particles" )
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/disable_death.png", "disable_death", "No Death" )
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/enable_logging.png", "spell_logging", "Spell Logging" )
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/quick_spell_picker.png", "quick_spell_picker", "Quick Spell Picker" )
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/2gcedge.png", "force_2gcedge", "2GC + Edge" );
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/no_recoil.png", "no_recoil", "No Recoil" );
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/unlimited_spells.png", "unlimited_spells", "Unlimited Spells" );
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/fxcrit.png", "force_fxcrit", "Force Crit on X" );
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/damage_info.png", "damage_info", "Damage Info" );
                        do_flag_toggle_image_button( "mods/spell_lab/files/gui/buttons/edit_wands_everywhere.png", "edit_wands_everywhere", "Edit Wands Everywhere" );
                    GuiLayoutEnd( gui );
                end

                if is_picker_active( PICKERS.WandBox ) then
                    GuiLayoutBeginHorizontal( gui, 45, 2 );
                        if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/save_wand.png" ) then
                            local held_wand = get_held_wand();
                            if held_wand then
                                local ability = EntityGetFirstComponentIncludingDisabled( held_wand, "AbilityComponent" );
                                if ability then
                                    local data = {};
                                    data.stats = ability_component_get_stats( ability );
                                    data.absolute_actions, data.permanent_actions = wand_get_actions_absolute( held_wand );
                                    table.insert(saved_wands, data );
                                    serialize_saved_wands();
                                end
                            end
                        end
                        GuiTooltip( gui, "Save Held Wand", "" );

                        if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/delete_wand.png" ) then
                            if selected_wand_box_index > #preset_wands then
                                table.remove( saved_wands, selected_wand_box_index - #preset_wands );
                                serialize_saved_wands();
                            end
                        end
                        GuiTooltip( gui, "Delete Selected Wand", "" );

                        if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/load_into_wand.png" ) then
                            local combined_wands = {};
                            for k,v in pairs( preset_wands ) do table.insert( combined_wands, v ); end
                            for k,v in pairs( saved_wands ) do table.insert( combined_wands, v ); end
                            local saved_wand = combined_wands[selected_wand_box_index];
                            if saved_wand then
                                local held_wand = get_held_wand();
                                if held_wand then
                                    initialize_wand( held_wand, saved_wand );
                                    force_refresh_held_wands();
                                end
                            end
                        end
                        GuiTooltip( gui, "Load Selected Wand Into Held Wand", "" );

                        if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/load_into_world.png" ) then
                            local combined_wands = {};
                            for k,v in pairs( preset_wands ) do table.insert( combined_wands, v ); end
                            for k,v in pairs( saved_wands ) do table.insert( combined_wands, v ); end
                            local saved_wand = combined_wands[selected_wand_box_index];
                            if saved_wand then
                                local x, y = get_player_or_camera_position();
                                local wand = EntityLoad( "data/entities/items/wand_level_03.xml", x, y );
                                if wand then initialize_wand( wand, saved_wand ); end
                            end
                        end
                        GuiTooltip( gui, "Load Selected Wand Into World", "" );
                    GuiLayoutEnd( gui );
                end

                if ModSettingGet( "spell_lab.damage_info" ) == true then
                    GuiLayoutAddVerticalSpacing( gui, 5 );
                    local screen_width, screen_height = GuiGetScreenDimensions( gui );
                    local highest_dps = GlobalsGetValue( "spell_lab_recent_highest_dps","" );
                    if #highest_dps > 0 then
                        GuiTextCentered( gui, screen_width * 0.5, 0, "Damage Per Second: "..highest_dps );
                    end

                    local total_damage = GlobalsGetValue( "spell_lab_recent_total_damage","" );
                    if #total_damage > 0 then
                        GuiTextCentered( gui, screen_width * 0.5, 0, "Total Damage: "..total_damage );
                    end

                    local player_projectiles = EntityGetWithTag("projectile_player") or {};
                    local highest_projectile_damage = 0;
                    local highest_damage_projectile = nil;
                    local total_projectile_damage = 0;
                    local total_projectiles = #player_projectiles;
                    for k,v in pairs( player_projectiles ) do
                        local projectile = EntityGetFirstComponent( v, "ProjectileComponent" );
                        if projectile then
                            local damage = ComponentGetValue2( projectile, "damage" ) * 25;
                            if damage > highest_projectile_damage then
                                highest_damage_projectile = v;
                                highest_projectile_damage = damage;
                            end
                            total_projectile_damage = total_projectile_damage + damage;
                        end
                    end
                    if highest_damage_projectile ~= nil then
                        local esx, esy = get_screen_position( EntityGetTransform( highest_damage_projectile ) );
                        GuiOptionsAddForNextWidget( gui, GUI_OPTION.Layout_NoLayouting );
                        GuiTextCentered( gui, esx, esy, thousands_separator( math.floor( highest_projectile_damage ) ) );
                    end
                    GuiTextCentered( gui, screen_width * 0.5, 0, "Total Projectile Damage: "..thousands_separator( math.floor( total_projectile_damage ) ) );
                    GuiTextCentered( gui, screen_width * 0.5, 0, "Total Projectiles: "..thousands_separator( math.floor( total_projectiles ) ) );
                end
            GuiLayoutEnd( gui );

            GuiLayoutBeginVertical( gui, 0, 92 );
                local held_wand = get_held_wand();
                    if held_wand and held_wand ~= 0 then
                        local max_actions = EntityGetWandCapacity( held_wand );
                        local held_wand_actions,held_wand_permanent_actions = wand_get_actions_absolute( held_wand );
                        local width = (max_actions + #held_wand_permanent_actions) * ( 20 + 2 );
                        local screen_width, screen_height = GuiGetScreenDimensions( gui );
                        GuiLayoutBeginHorizontal( gui, (screen_width - width) / screen_width / 2 * 100, 0 );
                            do_action_list( held_wand_actions, held_wand_permanent_actions, max_actions, held_wand );
                        GuiLayoutEnd( gui );
                    end
            GuiLayoutEnd( gui );

            if is_picker_active( PICKERS.WandPicker ) then
                GuiBeginAutoBox( gui );
                    GuiLayoutBeginVertical( gui, 5, 16 );
                        GuiLayoutBeginHorizontal( gui, 0, 0 );
                            GuiLayoutBeginVertical( gui, 0, 0 );
                                for _,stat in ipairs( wand_stats ) do
                                    GuiLayoutBeginHorizontal( gui, 0, 0 );
                                        local width = GuiGetTextDimensions( gui, stat.label );
                                        GuiColorSetForNextWidget( gui, 0.811, 0.811, 0.811, 1.0 );
                                        GuiText( gui, 60 - width, 0, stat.label );
                                        local left_click,right_click,hover,x,y,width,height,draw_x,draw_y = previous_data( gui );
                                        GuiOptionsAddForNextWidget( gui, GUI_OPTION.Layout_NoLayouting );
                                        GuiColorSetForNextWidget( gui, 1.0, 0.75, 0.5, 1.0 );
                                        GuiText( gui, x + width + 10, y, stat.text_callback( stat.current ) );
                                        stat.current = stat.value_callback( GuiSlider( gui, next_id(), 50, 0, " ", stat.current, stat.min, stat.max, stat.default, 1.0, stat.formatter, 120 ) );
                                    GuiLayoutEnd( gui );
                                end
                                GuiLayoutBeginHorizontal( gui, 0, 2 );
                                    if GuiButton( gui, 0, 0, "[Spawn]", next_id() ) then
                                        if player then
                                            local x, y = EntityGetTransform( player );
                                            local wand = EntityLoad( "data/entities/items/wand_level_01.xml", x, y );
                                            local wand_data = {
                                                stats = {}
                                            };
                                            for _,stat_data in pairs( wand_stats ) do
                                                wand_data.stats[stat_data.stat] = stat_data.current;
                                            end
                                            wand_data.stats["shuffle_deck_when_empty"] = wand_data.stats["shuffle_deck_when_empty"] == 1 and true or false;
                                            wand_clear_actions( wand );
                                            initialize_wand( wand, wand_data );
                                        end
                                    end
                                    if GuiButton( gui, 8, 0, "[Copy Held Wand]", next_id() ) then
                                        local held_wand = get_held_wand();
                                        if held_wand then
                                            local stats = wand_get_stats( held_wand );
                                            if stats ~= nil then
                                                for k,v in pairs( stats ) do
                                                    for _,stat_data in pairs( wand_stats ) do
                                                        if stat_data.stat == k then
                                                            stat_data.current = v;
                                                            break;
                                                        end
                                                    end
                                                end
                                            end
                                        end
                                    end
                                    if GuiButton( gui, 8, 0, "[Update Held Wand]", next_id() ) then
                                        local player = EntityGetWithTag( "player_unit" )[1];
                                        if player then
                                            local held_wand = get_entity_held_or_random_wand( player, false );
                                            if held_wand then
                                                local x, y = EntityGetTransform( player );
                                                local wand_data = {
                                                    stats = {}
                                                };
                                                for _,stat_data in pairs( wand_stats ) do
                                                    wand_data.stats[stat_data.stat] = stat_data.current;
                                                end
                                                wand_data.stats["shuffle_deck_when_empty"] = wand_data.stats["shuffle_deck_when_empty"] == 1 and true or false;
                                                initialize_wand( held_wand, wand_data, false );
                                            end
                                        end
                                    end
                                GuiLayoutEnd( gui );
                            GuiLayoutEnd( gui );
                        GuiLayoutEnd( gui );
                    GuiLayoutEnd( gui );
                    GuiZSetForNextWidget( gui, 1 );
                GuiEndAutoBoxNinePiece( gui );
            end

            if is_picker_active( PICKERS.SpellPicker ) then
                local adjusted_actions = sorted_actions[filter_type];
                if adjusted_actions and #adjusted_actions > 0 then
                    if current_challenge ~= nil then
                        adjusted_actions = {};
                        for _,action_id in pairs( current_challenge.available_actions ) do
                            local action = action_data[action_id];
                            table.insert( adjusted_actions, action );
                        end
                    end
                    local adjusted_columns = 8;
                    
                    GuiLayoutBeginVertical( gui, 5, 16 );
                        local scroll_id = next_id(9) + filter_type;
                        local rows = math.ceil( #adjusted_actions / adjusted_columns );
                        GuiBeginScrollContainer( gui, scroll_id, 0, 0, 174, math.min( rows * 20, 160 ) );
                            GuiLayoutBeginVertical( gui, 0, 0 );
                                    local action_index = 1;
                                    local action = adjusted_actions[action_index];
                                    while action ~= nil do
                                        GuiLayoutBeginHorizontal( gui, 0, 0 );
                                            for x=1,adjusted_columns do
                                                local action = adjusted_actions[action_index];
                                                if action then
                                                    do_action_button( action.id, 0, 0, false, function( left_click, right_click )
                                                        acquire_action( action.id, selected_acton_index, right_click );
                                                    end, true );
                                                end
                                                action_index = action_index + 1;
                                            end
                                            action = adjusted_actions[action_index];
                                        GuiLayoutEnd( gui );
                                        GuiLayoutAddVerticalSpacing( gui, -2 );
                                        --GuiLayoutAddHorizontalSpacing( gui, -2 );
                                    end
                            GuiLayoutEnd( gui );
                        GuiEndScrollContainer( gui );
                    GuiLayoutEnd( gui );
                end

                if current_challenge == nil then
                    GuiLayoutBeginVertical( gui, 1, 16 );
                        for i=0,7 do
                            if filter_type ~= i then GuiOptionsAddForNextWidget( gui, GUI_OPTION.DrawSemiTransparent ); end
                            GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/type_filter_"..i..".png" );
                            local left_click,right_click,hover,_,_,width = previous_data( gui );
                            GuiTooltip( gui, type_text[i], "" );
                            if hover then filter_type = i; end
                        end
                    GuiLayoutEnd( gui );
                end
            end

            if is_picker_active( PICKERS.ChallengePicker ) then
                local challenge_index = 1;
                GuiLayoutBeginVertical( gui, 5, 16 );
                    for y=1,10 do -- MAGIC
                        GuiLayoutBeginHorizontal( gui, 0, 0 );
                            for x=1,10 do -- MAGIC
                                local challenge = challenges[challenge_index];
                                if challenge == nil then break; end
                                local base_sprite = "challenge";
                                if challenge.challenge then base_sprite = "trophy"; end
                                local sprite = "mods/spell_lab/files/gui/buttons/"..base_sprite..".png";
                                if ModSettingGet( "spell_lab."..challenge.flag ) == true then
                                    sprite = "mods/spell_lab/files/gui/buttons/"..base_sprite.."_complete.png";
                                end
                                if GuiImageButton( gui, next_id(), 0, 0, "", sprite ) then
                                    for k,v in pairs( EntityGetWithTag( "spell_lab_challenge_cleanup" ) or {} ) do EntityKill( v ); end
                                    local player = EntityGetWithTag( "player_unit" )[1];
                                    if player then
                                        change_picker( PICKERS.SpellPicker );
                                        challenge.initialization_callback( challenge, player );
                                        local ex, ey = EntityGetTransform( player );
                                        for k,v in pairs( EntityGetWithTag("projectile_player" ) or {} ) do
                                            EntityKill( v );
                                        end
                                        local visualizer = EntityLoad( "data/entities/buildings/workshop_spell_visualizer.xml", ex - 78, ey - 50 );
                                        EntityAddTag( visualizer, "spell_lab_challenge_cleanup" );
                                        current_challenge = challenge;
                                        selected_acton_index = 0;
                                    end
                                end
                                
                                do_custom_tooltip( function()
                                    GuiLayoutBeginVertical( gui, 0, 0 );
                                        GuiText( gui, 0, 0, challenge.name );
                                        GuiColorSetForNextWidget( gui, 0.5, 0.5, 0.5, 1.0 );
                                        GuiText( gui, 0, 0, challenge.description );
                                        GuiLayoutBeginHorizontal( gui, 0, 0 );
                                            do_flat_action_list( challenge.available_actions );
                                        GuiLayoutEnd( gui );
                                    GuiLayoutEnd( gui );
                                end );
                                GuiZSet( gui, 0 );

                                --GuiTooltip( gui, challenge.name, challenge.description );
                                challenge_index = challenge_index + 1;
                            end
                        GuiLayoutEnd( gui );
                        GuiLayoutAddVerticalSpacing( gui, -2 );
                    end
                GuiLayoutEnd( gui );
            end

            if is_picker_active( PICKERS.WandBox ) then
                GuiLayoutBeginVertical( gui, 5, 16 );
                    local combined_wands = {};
                    for k,v in pairs( preset_wands ) do table.insert( combined_wands, v ); end
                    for k,v in pairs( saved_wands ) do table.insert( combined_wands, v ); end
                    local wand_index = 1;
                    for y=1,10 do
                        GuiLayoutBeginHorizontal( gui, 0, 0 );
                            for x=1,10 do
                                local saved_wand = combined_wands[wand_index];
                                if saved_wand == nil then break; end
                                local spell_box_suffix = "";
                                if selected_wand_box_index == wand_index then spell_box_suffix = spell_box_suffix .. "_active"; end
                                if GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/spell_box"..spell_box_suffix..".png" ) then
                                    if selected_wand_box_index == wand_index then
                                        selected_wand_box_index = 0;
                                    else
                                        selected_wand_box_index = wand_index;
                                    end
                                end
                                local left_click,right_click,hover,x,y,width,height,draw_x,draw_y = previous_data( gui );
                                do_custom_tooltip( function() 
                                    local capacity = 0;
                                    for k,v in pairs( saved_wand.absolute_actions ) do capacity = capacity + 1; end
                                    GuiLayoutBeginHorizontal( gui, 0, 0 );
                                        do_simple_action_list( saved_wand.absolute_actions, saved_wand.permanent_actions, capacity );
                                    GuiLayoutEnd( gui );
                                end );
                                GuiZSet( gui, 0 );
                                if saved_wand then
                                    local dynamic_wand_data = wand_get_dynamic_wand_data_from_stats( saved_wand.stats );
                                    if dynamic_wand_data then
                                        local image_width, image_height = GuiGetImageDimensions( gui, dynamic_wand_data.file );
                                        
                                        GuiZSetForNextWidget( gui, -1 );
                                        GuiOptionsAddForNextWidget( gui, GUI_OPTION.Layout_NoLayouting );
                                        GuiImage( gui, next_id(), x + ( 20 - image_height ) / 2.5, y + 10, dynamic_wand_data.file, 1.0, 1.0, 1.0, -60 / 180 * math.pi );
                                    end
                                end
                                wand_index = wand_index + 1;
                            end
                        GuiLayoutEnd( gui );
                        GuiLayoutAddVerticalSpacing( gui, -2 );
                    end
                GuiLayoutEnd( gui );
            end

            if current_challenge then
                GlobalsSetValue("spell_lab_challenge_active","1");
                local player = EntityGetWithTag( "player_unit" )[1];
                if player then
                    if current_challenge:validation_callback( player ) then
                        GamePrintImportant("Challenge Complete","");
                        --AddFlagPersistent( current_challenge.flag );
                        ModSettingSet( "spell_lab."..current_challenge.flag, true );
                        current_challenge = nil;
                        change_picker( PICKERS.ChallengePicker );
                    else
                        local progress_texts = current_challenge:progress_callback( entity );
                        
                        GuiBeginAutoBox( gui, 0, 0 );
                            GuiLayoutBeginVertical( gui, 1, 60 );
                                GuiLayoutBeginVertical( gui, 0, 0 );
                                    GuiText( gui, 0, 0, current_challenge.name );
                                    GuiColorSetForNextWidget( gui, 0.5, 0.5, 0.5, 1.0 );
                                    GuiText( gui, 4, 0, word_wrap( current_challenge.description, 40 ) );
                                    for _,v in pairs( progress_texts or {} ) do
                                        GuiText( gui, 4, 0, v );
                                    end    
                                    GuiLayoutBeginHorizontal( gui, 0, 0 );
                                        for wand_index,wand in ipairs( current_challenge.loadout.wands or {} ) do
                                            GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/challenge_wand.png" );
                                            do_custom_tooltip( function()
                                                GuiText( gui, 0, 0, "Challenge Wand ".. wand_index );
                                                do_wand_stats( gui, wand.stats );
                                            end, -30 );
                                            GuiZSet( gui, 0 );
                                        end
                                        for _,hint in pairs( current_challenge.hints or {} ) do
                                            GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/hint.png" );
                                            GuiTooltip( gui, word_wrap( hint ), "" );
                                        end
                                        for _,example in pairs( current_challenge.examples or {} ) do
                                            GuiImageButton( gui, next_id(), 0, 0, "", "mods/spell_lab/files/gui/buttons/example.png" );
                                            do_custom_tooltip( function()
                                                GuiLayoutBeginHorizontal( gui, 0, 0 );
                                                    do_flat_action_list( example );
                                                GuiLayoutEnd( gui );
                                            end );
                                            GuiZSet( gui, 0 );
                                        end
                                    GuiLayoutEnd( gui );
                                GuiLayoutEnd( gui );
                            GuiLayoutEnd( gui );

                            GuiZSetForNextWidget( gui, 1 );
                        GuiEndAutoBoxNinePiece( gui );
                    end
                end
            else
                GlobalsSetValue("spell_lab_challenge_active","0");
            end
        end

        local mod_button_reservation = tonumber( GlobalsGetValue( "spell_lab_mod_button_reservation", "0" ) );
        local current_button_reservation = tonumber( GlobalsGetValue( "mod_button_tr_current", "0" ) );
        if current_button_reservation > mod_button_reservation then
            current_button_reservation = mod_button_reservation;
        elseif current_button_reservation < mod_button_reservation then
            current_button_reservation = math.max( 0, mod_button_reservation + (current_button_reservation - mod_button_reservation ) );
        else
            current_button_reservation = mod_button_reservation;
        end
        GlobalsSetValue( "mod_button_tr_current", tostring( current_button_reservation + 15 ) );
        local player = EntityGetWithTag( "player_unit" )[1];
        if player then
            local platform_shooter_player = EntityGetFirstComponent( player, "PlatformShooterPlayerComponent" );
            if platform_shooter_player then
                local is_gamepad = ComponentGetValue2( platform_shooter_player, "mHasGamepadControlsPrev" );
                if is_gamepad then
                    GuiOptionsAddForNextWidget( gui, GUI_OPTION.NonInteractive );
                    GuiOptionsAddForNextWidget( gui, GUI_OPTION.AlwaysClickable );
                end
            end
        end
        if GuiImageButton( gui, next_id(), screen_width - 14 - current_button_reservation, 2, "", "mods/spell_lab/files/gui/wrench.png" ) then
            is_panel_open = not is_panel_open;
        end
        local text = (is_panel_open == true and "Hide" or "Show").." Spell Lab";
        local width = GuiGetTextDimensions( gui, text );
        do_custom_tooltip( function()
            GuiText( gui, 0, 0, text );
            GuiColorSetForNextWidget( gui, 0.5, 0.5, 0.5, 1.0 );
            GuiText( gui, 0, 0, version );
        end, -100, -width - 24, 10 );
        GuiIdPop( gui );
    end
    print("[spell lab] done setting up GUI");
end

do_gui();